﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CreateTextRes
{
    class Resources
    {
        static String backupFileExt = ".CreateTextResBackup";


        /* 9-12 bit LZW encoding */
        public static byte[] decrypt1(byte[] src, int length, int complength)
        /* Doesn't do length checking yet */
        {
            /* Theory: Considering the input as a bit stream, we get a series of
            ** 9 bit elements in the beginning. Every one of them is a 'token'
            ** and either represents a literal (if < 0x100), or a link to a previous
            ** token (tokens start at 0x102, because 0x101 is the end-of-stream
            ** indicator and 0x100 is used to reset the bit stream decoder).
            ** If it's a link, the indicated token and the character following it are
            ** placed into the output stream. Note that the 'indicated token' may
            ** very well consist of a link-token-plus-literal construct again, so
            ** it's possible to represent strings longer than 2 recursively.
            ** If the maximum number of tokens has been reached, the bit length is
            ** increased by one, up to a maximum of 12 bits.
            ** This implementation remembers the position each token was print to in
            ** the output array, and the length of this token. This method should
            ** be faster than the recursive approach.
            */

            byte[] dest = new byte[length];
            //Console.WriteLine(src.Length);

            ushort bitlen = 9; /* no. of bits to read (max. 12) */
            ushort bitmask = 0x01ff;
            ushort bitctr = 0; /* current bit position */
            ushort bytectr = 0; /* current byte position */
            ushort token; /* The last received value */
            ushort maxtoken = 0x200; /* The biggest token */

            ushort[] tokenlist = new ushort[4096]; /* pointers to dest[] */
            ushort[] tokenlengthlist = new ushort[4096]; /* char length of each token */
            ushort tokenctr = 0x102; /* no. of registered tokens (starts here)*/

            ushort tokenlastlength = 0;

            ushort destctr = 0;

            while (bytectr < complength)
            {

                ushort tokenmaker = (ushort)(src[bytectr++] >> bitctr);
                if (bytectr < complength)
                    tokenmaker |= (ushort)(src[bytectr] << (8 - bitctr));
                if (bytectr + 1 < complength)
                    tokenmaker |= (ushort)(src[bytectr + 1] << (16 - bitctr));

                token = (ushort)(tokenmaker & bitmask);

                bitctr += (ushort)(bitlen - 8);

                while (bitctr >= 8)
                {
                    bitctr -= 8;
                    bytectr++;
                }

                if (token == 0x101) return dest; /* terminator */
                if (token == 0x100)
                { /* reset command */
                    maxtoken = 0x200;
                    bitlen = 9;
                    bitmask = 0x01ff;
                    tokenctr = 0x0102;
                }
                else
                {

                    {
                        int i;

                        if (token > 0xff)
                        {
                            if (token >= tokenctr)
                            {
                                //fprintf(stderr, "decrypt1: Bad token %x!\n", token);

                                /* Well this is really bad  */
                                /* May be it should throw something like SCI_ERROR_DECOMPRESSION_INSANE */
                            }
                            else
                            {
                                tokenlastlength = (ushort)(tokenlengthlist[token] + 1);
                                if (destctr + tokenlastlength > length)
                                {
                                    /* For me this seems a normal situation, It's necessary to handle it*/
                                    //printf ("decrypt1: Trying to write beyond the end of array(len=%d, destctr=%d, tok_len=%d)!\n",
                                    //    length, destctr, tokenlastlength);

                                    i = 0;
                                    for (; destctr < length; destctr++)
                                    {
                                        dest[destctr++] = dest[tokenlist[token] + i];
                                        i++;
                                    }
                                }
                                else
                                    for (i = 0; i < tokenlastlength; i++)
                                    {
                                        dest[destctr++] = dest[tokenlist[token] + i];
                                    }
                            }
                        }
                        else
                        {
                            tokenlastlength = 1;
                            if (destctr >= length)
                            {
                                Console.WriteLine("decrypt1: Try to write single byte beyond end of array!\n");
                            }
                            else
                                dest[destctr++] = (byte)token;
                        }

                    }

                    if (tokenctr == maxtoken)
                    {
                        if (bitlen < 12)
                        {
                            bitlen++;
                            bitmask <<= 1;
                            bitmask |= 1;
                            maxtoken <<= 1;
                        }
                        else continue; /* no further tokens allowed */
                    }
                    tokenlist[tokenctr] = (ushort)(destctr - tokenlastlength);
                    tokenlengthlist[tokenctr++] = tokenlastlength;
                }
            }
            return dest;
        }


        public static void Extract(string inputResourceFile, string extractPath, bool preserveOrder)
        {
            /* Note: preserveOrder will append the relative position of the patch resource within the RESOURCE.001 file 
             * to the front on the file name, allowing repacking in the same order to be done
             */
            byte[] _bytes = null;
            String hexResNum = "";
            int resLengthComp = 0;
            int resLengthDecomp = 0;
            int resType = 0;
            int resNum = -1;  // init for patch resource
            int compType = 0;
            String currentRes = "";
            int extractCount = 0;

            if (Directory.Exists(extractPath))
            {
                // delete the extracted resource folder
                System.IO.DirectoryInfo downloadedMessageInfo = new DirectoryInfo(extractPath);

                foreach (FileInfo file in downloadedMessageInfo.GetFiles())
                {
                    file.Delete();
                }
                foreach (DirectoryInfo dir in downloadedMessageInfo.GetDirectories())
                {
                    dir.Delete(true);
                }
            }



            try
            {
                _bytes = File.ReadAllBytes(inputResourceFile);
            }
            catch
            {
                Console.WriteLine("Could not open input file!");
                Environment.Exit(2);   // exit with error
            }
            String hexString = HexUtilities.BytesToHexString(_bytes);

            Console.WriteLine("Extracting resources from " + inputResourceFile);

            // This is a RESOURCE.XXX file
            while (hexString.Length > 0)
            {

                // Get the first 16 characters (8 bytes) of the header
                hexResNum = hexString.Substring(0, 4);
                resLengthComp = Convert.ToInt32(hexString.Substring(4, 2), 16) + Convert.ToInt32(hexString.Substring(6, 2), 16) * 256 + 4;
                resLengthDecomp = Convert.ToInt32(hexString.Substring(8, 2), 16) + Convert.ToInt32(hexString.Substring(10, 2), 16) * 256 + 4;
                resType = Convert.ToInt32(HexUtilities.hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(0, 5), 2);
                resNum = Convert.ToInt32(hexString.Substring(0, 2), 16)
                            + Convert.ToInt32(HexUtilities.hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(5, 3), 2) * 256;

                compType = Convert.ToInt32(hexString.Substring(12, 2), 16);

                // Split into two pieces, this resource and the rest of the file
                try
                {
                    currentRes = hexString.Substring(16, resLengthComp * 2 - 16);
                    hexString = hexString.Substring(resLengthComp * 2);
                }
                catch
                {
                    Console.WriteLine("Non SCI0 resource or resource patch file.  Exiting.");
                    Environment.Exit(2); // Exit with error
                }

                // LZW compression
                if (compType == 1)
                {
                    byte[] destination = decrypt1(HexUtilities.HexStringToBytes(currentRes), resLengthDecomp - 4, resLengthComp - 8);
                    currentRes = HexUtilities.BytesToHexString(destination);
                }

                // Huffman
                if (compType == 2)
                {
                    Console.WriteLine("Skipping Huffman compressed resource (not currently supported).");
                }


                // write the uncompressed patch resource to our working dir
                byte[] outputBytes = HexUtilities.HexStringToBytes(currentRes);
                string extractFileName = null;
                String fileExt = resNum.ToString();

                // left pad the resource numbers to 3 digits
                fileExt = fileExt.PadLeft(3, '0');

                switch (resType)
                {
                    case 0: extractFileName = "VIEW." + fileExt; break;
                    case 1: extractFileName = "PIC." + fileExt; break;
                    case 2: extractFileName = "SCRIPT." + fileExt; break;
                    case 3: extractFileName = "TEXT." + fileExt; break;
                    case 4: extractFileName = "SOUND." + fileExt; break;
                    case 6: extractFileName = "VOCAB." + fileExt; break;
                    case 7: extractFileName = "FONT." + fileExt; break;
                    case 8: extractFileName = "CURSOR." + fileExt; break;
                    case 9: extractFileName = "PATCH." + fileExt; break;
                    default:
                        Console.WriteLine("Something went wrong.  Cannot determine resource type.");
                        Environment.Exit(2);   // exit with error
                        break;
                }

                if (preserveOrder)
                {
                    extractFileName = extractCount.ToString().PadLeft(3, '0') + '_' + extractFileName;
                }

                extractFileName = Path.Combine(extractPath, extractFileName);

                try
                {
                    System.IO.FileInfo extractFile = new System.IO.FileInfo(extractFileName);
                    extractFile.Directory.Create(); // If the directory already exists, this method does nothing.
                    FileStream fsExtract = new FileStream(extractFileName, FileMode.Create, FileAccess.Write);

                    // write out the 2 byte header
                    fsExtract.Write(HexUtilities.HexStringToBytes("8" + resType + "00"), 0, 2);

                    // write out the resource
                    fsExtract.Write(outputBytes, 0, outputBytes.Length);
                    fsExtract.Close();
                    extractCount++;
                }
                catch
                {
                    Console.WriteLine("Could not create output file!");
                    Environment.Exit(2);
                }
            }



        }

        public static void Pack(string inputDirectory, string outputResourceFile)
        {
            int resTypeDec = 0;
            string[] fileEntries = Directory.GetFiles(inputDirectory);
            FileStream fsResourceFile = null;

            string destPath = Path.GetDirectoryName(outputResourceFile);
            string destFile = Path.GetFileName(outputResourceFile);


            // delete the backup file if it exists
            if (File.Exists(Path.Combine(destPath, destFile + backupFileExt)))
            {
                File.Delete(Path.Combine(destPath, destFile + backupFileExt));
            }

            // copy the original script file to the backup
            File.Copy(Path.Combine(destPath, destFile), Path.Combine(destPath, destFile + backupFileExt));


            // delete the new resource file if it exists
            if (File.Exists(outputResourceFile))
            {
                File.Delete(outputResourceFile);
            }

            // open the file for writing
            try
            {
                System.IO.FileInfo resourceFile = new System.IO.FileInfo(outputResourceFile);
                resourceFile.Directory.Create(); // If the directory already exists, this method does nothing.
                fsResourceFile = new FileStream(outputResourceFile, FileMode.Create, FileAccess.Write);
            }
            catch
            {
                Console.WriteLine("Could not create output resource file!");
                Environment.Exit(2);
            }


            foreach (string fileName in fileEntries)
            {
                string fileExt = Path.GetExtension(fileName);
                string fileNameNoExt = Path.GetFileNameWithoutExtension(fileName);
                string switchVal = null;

                if (fileNameNoExt.Contains("_"))
                {
                    switchVal = fileNameNoExt.ToUpper().Substring(4);
                }
                else
                { 
                    switchVal = fileNameNoExt.ToUpper();
                }

                switch (switchVal)
                {
                    case "VIEW": resTypeDec = 0; break;
                    case "PIC": resTypeDec = 1; break;
                    case "SCRIPT": resTypeDec = 2; break;
                    case "TEXT": resTypeDec = 3; break;
                    case "SOUND": resTypeDec = 4; break;
                    case "VOCAB": resTypeDec = 6; break;
                    case "FONT": resTypeDec = 7; break;
                    case "CURSOR": resTypeDec = 8; break;
                    case "PATCH": resTypeDec = 9; break;
                    default:
                        Console.WriteLine("Something went wrong.  Cannot determine resource type.");
                        Environment.Exit(2);   // exit with error
                        break;
                }

                string resNumBinary = Convert.ToString(Convert.ToInt32(fileExt.Substring(1)),2).PadLeft(11,'0');
                resNumBinary = resNumBinary.Substring(resNumBinary.Length - 11);

                string resTypeBinary = Convert.ToString(resTypeDec, 2).PadLeft(8,'0');
                resTypeBinary = resTypeBinary.Substring(resTypeBinary.Length - 5);

                string bin1 = resNumBinary.Substring(3,8);
	            string bin2 = resTypeBinary + resNumBinary.Substring(0, 3);

                // read the whole file in
                byte[] _bytes = null;

                try
                {
                    _bytes = File.ReadAllBytes(fileName);
                }
                catch
                {
                    Console.WriteLine("Could not open input file!");
                    Environment.Exit(2);   // exit with error
                }
                String hexString = HexUtilities.BytesToHexString(_bytes);

                // strip off the first two bytes from the file (8X00)
                hexString = hexString.Substring(4);

                int resourceLength = hexString.Length;

                // compressed size
                int compressedLength = (resourceLength / 2) + 4;
                int compressedSize1 = compressedLength % 256;
                int compressedSize2 = compressedLength / 256;

                // decompressed size
                int decompressedLength = resourceLength / 2;
                int decompressedSize1 = decompressedLength % 256;
                int decompressedSize2 = decompressedLength / 256;

                hexString = String.Format("{0:X2}", Convert.ToInt32(bin1, 2))
                          + String.Format("{0:X2}", Convert.ToInt32(bin2, 2))
                          + compressedSize1.ToString("X").PadLeft(2,'0') + compressedSize2.ToString("X").PadLeft(2,'0')
                          + decompressedSize1.ToString("X").PadLeft(2, '0') + decompressedSize2.ToString("X").PadLeft(2, '0')
                          + "0000" + hexString;

                // write out the file
                byte[] resourceBytes = HexUtilities.HexStringToBytes(hexString);

                fsResourceFile.Write(resourceBytes, 0, resourceBytes.Length);
            }
            fsResourceFile.Close();
        }

        public static void CreateMapFile(string inputResourceFile, string outputMapFile)
        {
            byte[] _bytes = null;
            String hexResNum = "";
            int resLengthComp = 0;
            int resLengthDecomp = 0;
            int resType = 0;
            int resNum = -1;  // init for patch resource
            int compType = 0;
            int filePos = 0;

            String currentRes = "";
            int resourceFileNum = Convert.ToInt32(Path.GetExtension(inputResourceFile).Substring(1));

            String resourceFileNumBin = (Convert.ToString(resourceFileNum, 2) + "00").PadLeft(6, '0');
            resourceFileNumBin = resourceFileNumBin.Substring(resourceFileNumBin.Length - 6);

            String resourceFileNumHex =  String.Format("{0:X2}", Convert.ToInt32(resourceFileNumBin, 2));


            string destPath = Path.GetDirectoryName(outputMapFile);
            string destFile = Path.GetFileName(outputMapFile);


            // delete the backup file if it exists
            if (File.Exists(Path.Combine(destPath, destFile + backupFileExt)))
            {
                File.Delete(Path.Combine(destPath, destFile + backupFileExt));
            }

            // copy the original map file to the backup
            File.Copy(Path.Combine(destPath, destFile), Path.Combine(destPath, destFile + backupFileExt));

            // delete the new map file if it exists
            if (File.Exists(outputMapFile))
            {
                File.Delete(outputMapFile);
            }

            try
            {
                _bytes = File.ReadAllBytes(inputResourceFile);
            }
            catch
            {
                Console.WriteLine("Could not open input file!");
                Environment.Exit(2);   // exit with error
            }
            String hexString = HexUtilities.BytesToHexString(_bytes);


            Console.WriteLine("Recreating MAP file from " + inputResourceFile);

            // Create the new MAP file
            FileStream mapFile = new FileStream(outputMapFile, FileMode.Create, FileAccess.Write);

            // This is a RESOURCE.XXX file
            while (hexString.Length > 0)
            {

                // Get the first 16 characters (8 bytes) of the header
                hexResNum = hexString.Substring(0, 4);
                resLengthComp = Convert.ToInt32(hexString.Substring(4, 2), 16) + Convert.ToInt32(hexString.Substring(6, 2), 16) * 256 + 4;
                resLengthDecomp = Convert.ToInt32(hexString.Substring(8, 2), 16) + Convert.ToInt32(hexString.Substring(10, 2), 16) * 256 + 4;
                resType = Convert.ToInt32(HexUtilities.hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(0, 5), 2);
                resNum = Convert.ToInt32(hexString.Substring(0, 2), 16)
                            + Convert.ToInt32(HexUtilities.hex2binary(hexString.Substring(2, 2)).PadLeft(8, '0').Substring(5, 3), 2) * 256;

                compType = Convert.ToInt32(hexString.Substring(12, 2), 16);

                // Split into two pieces, this resource and the rest of the file
                try
                {
                    currentRes = hexString.Substring(16, resLengthComp * 2 - 16);
                    hexString = hexString.Substring(resLengthComp * 2);
                }
                catch
                {
                    Console.WriteLine("Non SCI0 resource or resource patch file.  Exiting.");
                    Environment.Exit(2); // Exit with error
                }

                // LZW compression
                if (compType == 1)
                {
                    byte[] destination = decrypt1(HexUtilities.HexStringToBytes(currentRes), resLengthDecomp - 4, resLengthComp - 8);
                    currentRes = HexUtilities.BytesToHexString(destination);
                }

                // Huffman
                if (compType == 2)
                {
                    Console.WriteLine("Skipping Huffman compressed resource (not currently supported).");
                }

                String filePosHex = filePos.ToString("X").PadLeft(6, '0');

                // reverse the hex values
                filePosHex = filePosHex.Substring(4, 2) + filePosHex.Substring(2, 2) + filePosHex.Substring(0, 2);

                // append the MAP file data
                mapFile.Write(HexUtilities.HexStringToBytes(hexResNum), 0, 2);
                mapFile.Write(HexUtilities.HexStringToBytes(filePosHex), 0, 3);
                mapFile.Write(HexUtilities.HexStringToBytes(resourceFileNumHex), 0, 1);

                filePos = filePos + (currentRes.Length / 2) + 8;
            }

            // write file footer
            mapFile.Write(HexUtilities.HexStringToBytes("FFFFFFFFFFFF"), 0, 6);
            mapFile.Close();

        }

    }
}
